unit U_ChiSquare_RNG_Testing;
{Copyright  2013, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  shellAPI, StdCtrls, ExtCtrls, ComCtrls, mathslib, Spin;

type
  TForm1 = class(TForm)
    StaticText1: TStaticText;
    Panel1: TPanel;
    Memo1: TMemo;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    Memo2: TMemo;
    CHSQDemoBtn: TButton;
    Memo3: TMemo;
    Memo4: TMemo;
    DecilesBtn: TButton;
    DigitsBtn: TButton;
    KnuthBtn: TButton;
    QuantilesEdt: TSpinEdit;
    Label2: TLabel;
    TrialsEdt: TSpinEdit;
    Label3: TLabel;
    SamplesEdt: TSpinEdit;
    Label4: TLabel;
    procedure StaticText1Click(Sender: TObject);
    procedure CHSQDemoBtnClick(Sender: TObject);
    procedure DigitsBtnClick(Sender: TObject);
    procedure DecilesBtnClick(Sender: TObject);
    procedure KnuthBtnClick(Sender: TObject);
    procedure Memo4Change(Sender: TObject);
  private
    { Private declarations }
  public

end;

var
  Form1: TForm1;

implementation


{$R *.DFM}

 const
  nbins=10;
  npts=2000;
TYPE
   barray = ARRAY [1..10] of extended;
   bExtArray = array of extended;
   //bint64array = array of int64;
VAR
   glinext,glinextp : integer;
   glma : ARRAY [1..55] OF extended;
   chsq, chsq64, df, prob, prob64, x : extended;
   ibin,idum : integer;

{******************************************************}
{      "Numerical Recipes in Pascal" Includes          }
{******************************************************}
//{$I 'Modfile.pas'}
 TYPE double = extended;
         char12 = string[12];

    FUNCTION sngl(x:extended):extended;
    BEGIN
      sngl := x
    END;

    PROCEDURE glopen(VAR infile:text; filename:char12);
    BEGIN
      assign(infile,filename);
      reset(infile)
    END;
//{$I 'Gammln.pas'}
FUNCTION gammln(xx: extended): extended;
CONST
   stp = 2.50662827465;
   half = 0.5;
   one = 1.0;
   fpf = 5.5;
VAR
   x,tmp,ser: double;
   j: integer;
   cof: ARRAY [1..6] OF double;
BEGIN
   cof[1] := 76.18009173;
   cof[2] := -86.50532033;
   cof[3] := 24.01409822;
   cof[4] := -1.231739516;
   cof[5] := 0.120858003e-2;
   cof[6] := -0.536382e-5;
   x := xx-one;
   tmp := x+fpf;
   tmp := (x+half)*ln(tmp)-tmp;
   ser := one;
   FOR j := 1 TO 6 DO BEGIN
      x := x+one;
      ser := ser+cof[j]/x
   END;
   gammln := sngl(tmp+ln(stp*ser))
END;
//{$INCLUDE  'GCF.PAS'}
PROCEDURE gcf(a,x: extended; VAR gammcf,gln: extended);
LABEL 1;
CONST
   itmax=100;
   eps=3.0e-7;
VAR
   n: integer;
   gold,g,fac,b1,b0,anf,ana,an,a1,a0: extended;
BEGIN
   gln := gammln(a);
   gold := 0.0;
   a0 := 1.0;
   a1 := x;
   b0 := 0.0;
   b1 := 1.0;
   fac := 1.0;
   g:=0.0; {GDD - ensure g is initialized}
   FOR n := 1 TO itmax DO BEGIN
      an := 1.0*n;
      ana := an-a;
      a0 := (a1+a0*ana)*fac;
      b0 := (b1+b0*ana)*fac;
      anf := an*fac;
      a1 := x*a0+anf*a1;
      b1 := x*b0+anf*b1;
      IF (a1 <> 0.0) THEN BEGIN
         fac := 1.0/a1;
         g := b1*fac;
         IF (abs((g-gold)/g) < eps) THEN GOTO 1;
         gold := g
      END
   END;
   showmessage('pause in GCF - a too large, itmax too small'); //readln;                                                      *)adln;
1:   gammcf := exp(-x+a*ln(x)-gln)*g
END;
//{$INCLUDE 'GSER.PAS'}
PROCEDURE gser(a,x: extended; VAR gamser,gln: extended);
LABEL 1;
CONST
   itmax=100;
   eps=3.0e-7;
VAR
   n: integer;
   sum,del,ap: extended;
BEGIN
   gln := gammln(a);
   IF (x <= 0.0) THEN BEGIN
      IF (x < 0.0) THEN BEGIN
         writeln('pause in GSER - x less than 0'); readln
      END;
      gamser := 0.0
   END ELSE BEGIN
      ap := a;
      sum := 1.0/a;
      del := sum;
      FOR n := 1 TO itmax DO BEGIN
         ap := ap+1.0;
         del := del*x/ap;
         sum := sum+del;
         IF (abs(del) < abs(sum)*eps) THEN GOTO 1
      END;
      writeln('pause in GSER - a too large, itmax too small'); readln;
1:      gamser := sum*exp(-x+a*ln(x)-gln)
   END
END;
//{$INCLUDE 'GAMMQ.PAS'}
FUNCTION gammq(a,x: extended):extended;
VAR
   gamser,gln: extended;
BEGIN
   IF ((x < 0.0) OR (a <= 0.0)) THEN BEGIN
      showmessage('Pause in GAMMQ - invalid arguments');
   END;
   IF (x < a+1.0) THEN BEGIN
      gser(a,x,gamser,gln);
      gammq := 1.0-gamser
   END ELSE BEGIN
      gcf(a,x,gamser,gln);
      gammq := gamser
   END
END;
//{$INCLUDE 'RAN3.PAS'}
FUNCTION ran3(VAR idum: integer): extended;
(* Programs using RAN3 must declare the following variables
VAR
   glinext,glinextp: integer;
   glma: ARRAY [1..55] OF extended;
in the main routine. Machines with 4-byte integers can use the integer
implementation of this routine, substituting glma of type integer, the
commented CONST and VAR declarations, and the MOD function in the third
line after the BEGIN. *)
(* CONST
   mbig=1000000000;
   mseed=161803398;
   mz=0;
   fac=1.0e-9;
VAR
   i,ii,k,mj,mk: integer; *)
CONST
   mbig=4.0e6;
   mseed=1618033.0;
   mz=0.0;
   fac=2.5e-7; (* 1/mbig *)
VAR
   i,ii,k: integer;
   mj,mk: extended;
BEGIN
   IF (idum < 0) THEN BEGIN
      mj := mseed+idum;
      (* The following IF block is mj := mj MOD mbig; for extended variables. *)
      IF mj>=0.0 THEN mj := mj-mbig*trunc(mj/mbig)
         ELSE mj := mbig-abs(mj)+mbig*trunc(abs(mj)/mbig);
      glma[55] := mj;
      mk := 1;
      FOR i := 1 TO 54 DO BEGIN
         ii := 21*i MOD 55;
         glma[ii] := mk;
         mk := mj-mk;
         IF (mk < mz) THEN mk := mk+mbig;
         mj := glma[ii]
      END;
      FOR k := 1 TO 4 DO BEGIN
         FOR i := 1 TO 55 DO BEGIN
            glma[i] := glma[i]-glma[1+((i+30) MOD 55)];
            IF (glma[i] < mz) THEN glma[i] := glma[i]+mbig
         END
      END;
      glinext := 0;
      glinextp := 31;
      idum := 1
   END;
   glinext := glinext+1;
   IF (glinext = 56) THEN glinext := 1;
   glinextp := glinextp+1;
   IF (glinextp = 56) THEN glinextp := 1;
   mj := glma[glinext]-glma[glinextp];
   IF (mj < mz) THEN mj := mj+mbig;
   glma[glinext] := mj;
   ran3 := mj*fac
END;
//{$INCLUDE 'EXPDEV.PAS'}
FUNCTION expdev(VAR idum: integer): extended;
BEGIN
   expdev := -ln(ran3(idum))
END;

//{$INCLUDE 'CHSONE.PAS'}
PROCEDURE chsone(bins,ebins: barray; nbins,knstrn: integer;
       VAR df,chsq,prob: extended );
       (* Programs using routine CHSONE must define the types
TYPE
   barray = ARRAY [1..nbins] OF extended;
in the main routine.   *)
VAR
   j: integer;
BEGIN
   df := nbins-1-knstrn;
   chsq := 0.0;
   FOR j := 1 TO nbins DO BEGIN
      IF (ebins[j] <= 0.0) THEN BEGIN
         writeln('pause in CHSONE - bad expected number')
      END;
      chsq := chsq+sqr(bins[j]-ebins[j])/ebins[j]
   END;
   prob := gammq(0.5*df,0.5*chsq)
END;


PROCEDURE chsoneExt(bins,ebins: bExtArray; nbins,knstrn: integer;
       VAR df,chsq,prob: extended);
{ Programs using routine CHSONE must define the types
TYPE
   barray = ARRAY [1..nbins] OF extended;
in the main routine. }
VAR
   j: integer;
BEGIN
   df := nbins-1-knstrn;
   chsq := 0.0;
   FOR j := low(bins) TO high(bins) do BEGIN
      IF (ebins[j] <= 0.0) THEN BEGIN
         showmessage('Pause in CHSONE - bad expected number')
      END;
      chsq := chsq+sqr(bins[j]-ebins[j])/ebins[j]
   END;
   prob := gammq(0.5*df,0.5*chsq)
END;



{******************************************************}
{  End of "Numerical Recipes in Pascal" Includes       }
{******************************************************}


{************ CHSQDemoBtnClick **************}
 procedure TForm1.CHSQDemoBtnClick(Sender: TObject) ;
 var
   bins,ebins : barray;
   i,j:integer;
BEGIN
   idum := -15;
   FOR j := 1 to nbins DO
   BEGIN
      bins[j] := 0.0
   END;
   FOR i := 1 to npts DO
   BEGIN
      x := expdev(idum);
      ibin := trunc(x*nbins/3.0)+1;
      IF (ibin <= nbins) THEN
         bins[ibin] := bins[ibin]+1.0
   END;
   FOR i := 1 to nbins DO
   BEGIN
      ebins[i] := 3.0*npts/nbins*exp(-3.0*(i-0.5)/nbins);
   END;
   chsone(bins,ebins,nbins,-1,df,chsq,prob);

   with memo1.Lines do
   begin
     add('Expected      Observed');
     FOR i := 1 to nbins DO
     BEGIN
        add(format('%6.0f        %6.0f', [ebins[i],bins[i]]));
     END;
     add('');
     add(format('DF=%.0f',[df]));
     add('Probability = Chance that true Chi-Sq is greater than observed');
     add(format('Chi-squared: %10.4f',[chsq]));
     add(format('Probability: %10.4f', [prob]));
   end;
end;

var
KnuthExpected:array[0..10] of integer= (4,8,12,16,20,24,20,16,12,8,4);
KnuthObserved:array[0..10] of integer= (2,4,10,12,22,29,21,15,14,9,6);
KnuthObservedTooBad:array[0..10] of integer= (4,10,10,13,20,18,18,11,13,14,13);
KnuthObservedTooGood:array[0..10] of integer= (3,7,11,15,19,24,21,17,13,9,5);

{********** KnuthBtnClick **********}
procedure TForm1.KnuthBtnClick(Sender: TObject);
{Knuth test1}
var
  nbins:integer;
  bins1,bins2,bins3 : bExtArray;
  ebins: bExtArray;
  chsq1,chsq2,chsq3,prob1,prob2,prob3:extended;
  i:integer;
begin
  nbins:=11;
  setlength(bins1, nbins);
  setlength(bins2, nbins);
  setlength(bins3, nbins);
  setlength(ebins,nbins);
  for i:=0 to nbins-1 do
  BEGIN
     ebins[i] := knuthexpected[i];
     bins1[i]:= KnuthObserved[i];
     bins2[i]:= KnuthObservedTooBad[i];
     bins3[i]:= KnuthObservedTooGood[i];
  END;
  chsoneExt(bins1,ebins,nbins,0,df,chsq1,prob1);
  chsoneExt(bins2,ebins,nbins,0,df,chsq2,prob2);
  chsoneExt(bins3,ebins,nbins,0,df,chsq3,prob3);
  with memo1.lines do
  begin
    add('');
    add(' Expected    Observed    Large Values  Low Values');
    add('                            Test           Test') ;
    FOR i := 0 to nbins-1 DO add(format('%2.0f %15.0f %15.0f %15.0f   ',
                 [ebins[i],bins1[i], bins2[i], bins3[i]]));
    add('');
    add(format('DF=%.0f',[df]));
    add('Probability = Chance that drawing 11 samples from the expected distribution would have a Chi-Sq value >= the observed Chi-Sq.');
    add(format(' Chi-squared Observed:%7.4f, Probability %7.4f',[chsq1, prob1]));
    add(format(' Chi-squared Too High Demo: %7.4f, Probability %7.4f',[chsq2, prob2]));
    add(format(' Chi-squared Too Low Demo: %7.4f, Probability %7.4f',[chsq3, prob3]));
  end;
end;



procedure TForm1.Memo4Change(Sender: TObject);
begin

end;

{******** DigitsBtnClick ***********}
{RNG_Digits Tests}
procedure TForm1.DigitsBtnClick(Sender: TObject);

var
  nbins:integer;
  npts:integer;
  n:integer;
  bins : bExtArray;
  ebins: bExtArray;
  bins64, ebins64: bExtArray;
  chsqsum,chsqsum64, probsum, probsum64 :extended;
  nloops:integer;
  i,J, w, sum:integer;
  w64, sum64:int64;
begin
   screen.cursor:=crHourGlass;
   nbins:=10;
   npts:=SamplesEdt.value;
   nloops:=TrialsEdt.value;
   chsqsum:=0; chsqsum64:=0; probsum:=0; probsum64:=0;
   randomize;
   randomize64;
   setlength(bins, nbins);
   setlength(ebins,nbins);
   setlength(bins64,nbins);
   setlength(ebins64,nbins);
   for j:=1 to nloops do
   with memo1.Lines do
   begin

     FOR i := 0 to nbins-1 DO
     BEGIN
        bins[i] := 0;
        bins64[i]:=0;
        ebins[i]:= npts div nbins;
        ebins64[i]:=ebins[i];
     END;
     sum:=0;
     sum64:=0;
     FOR i := 1 to npts DO
     BEGIN

       {genrate and count distribution of random 1 digit numbers}
       n:=random(10);
       bins[n]:=bins[n]+1;
       n:=random64(10);
       bins64[n]:=bins64[n]+1;

      (*
         {Counting digit occurrences from larger random numbers would only work if
           we calculated the total number of digits from 0 to 1e10 -1 (or 1e18-1
           and divided by 10 to get expexted values fpor each digit,
           Even after accounti ng for this, 32 bit returns seem to be overloaded
           with zeros and ones.  Set aside for now }
      // w:=random(trunc(1e10));
      w:=random(100);
       while w>=10  do
       begin
         n:=w mod 10;
         bins[n]:=bins[n]+1;
         inc(sum);
         w:=w div 10;
       end;
       w64:=trunc(random64*1E18);
       while w64>=10  do
       begin
         n:=w64 mod 10;
         bins64[n]:=bins64[n]+1;
         inc(sum64);
         w64:=w64 div 10;
       end;
       *)
     END;
     (*
     for i:=0 to 9 do
     begin
       ebins[i]:=sum div 10;
       ebins64[i]:=sum64 div 10;
     end;
     *)
     chsoneExt(bins,ebins,nbins,0,df,chsq,prob);
     chsoneExt(bins64,ebins64 ,nbins,0,df,chsq64,prob64);
     chsqsum:=chsqsum+chsq; probsum:=probsum+prob;
     chsqsum64:=chsqsum64+chsq64; probsum64:=probsum64+prob64;
     if j<=10 then
     begin
       add('');
       add(format('Trial # %d (%.0n randon samples}',[j,0.0+npts]));
       add('33 bitExpected      32 bit Observed     64 bit observed');
        add('  32 bit    32 bit      64 bit    64 bit');
       add(' Expected  Observed    Expected  Observed');
       FOR i := 0 to nbins-1 DO
          add(format('%7.0f  %7.0f      %7.0f %7.0f', [ebins[i],bins[i], ebins64[i], bins64[i]]));
       (*
       FOR i := 0 to nbins-1 DO
       BEGIN
          add(format('%14.0f   %14.0f          %14.0f', [ebins[i],bins[i], bins64[i]]));
       END;
       *)
       add('');
       add(format('DF=%.0f',[df]));
       add(format('Probability = Chance that drawing %d samples from the expected '
          + 'distribution would have a Chi-Sq value >= the observed Chi-Sq.',[npts]));
       add(format('32 bit Chi-squared: %10.4f,    Probability %10.4f',[chsq, prob]));
       add(format('64 bit Chi-Squared  %10.4f,  Probability %10.4f', [chsq64, prob64]));
     end;
   end;
   with memo1.Lines do
   begin
     add('');
     add(format('Mean values for %d trials with %.0n samples per trial',[nloops,0.0+npts]));
     add(format('32 bit Chi-squared: %10.4f,    Probability %10.4f',[chsqsum/nloops, probsum/nloops]));
     add(format('64 bit Chi-Squared  %10.4f,  Probability %10.4f', [chsqsum64/nloops, probsum64/nloops]));
   end;
    screen.cursor:=crDefault;
end;



{************** DecileBtnClick ***********}
{RNG Deciles test}
procedure TForm1.DecilesBtnClick(Sender: TObject);
  var
  nbins:integer;
  npts:integer;
  n:integer;
  //n64:int64;
  bins,ebins : bExtArray;
  bins64 : bExtArray;
  ebins64 : bExtArray;
  chsqsum,chsqsum64, probsum, probsum64 :extended;
  nloops:integer;
  groupsize:integer;
  groupsize64:int64;
  i,j:integer;
begin
   screen.cursor:=crHourGlass;
   nbins:=QuantilesEdt.value;
   npts:=SamplesEdt.Value;
   nloops:=TrialsEdt.Value;
   chsqsum:=0; chsqsum64:=0; probsum:=0; probsum64:=0;
   //randseed:=123456;  {for testing}
   //randseed64:=123456;
   randomize;
   randomize64;
   setlength(bins, nbins);
   setlength(ebins,nbins);
   setlength(bins64,nbins);
   setlength(ebins64,nbins);
   groupsize:=high(integer) div nbins;
   groupsize64:=high(int64) div nbins;
   //groupsize64:=high(integer) div nbins;
   for j:=1 to nloops do
   with memo1.Lines do
   begin
     FOR i := 0 to nbins-1 DO
     BEGIN
        bins[i] := 0;
        bins64[i]:=0;
        ebins[i]:= npts div nbins;
        ebins64[i]:=npts div nbins;
     END;

     FOR i := 1 to npts DO
     BEGIN
       n:=trunc(random*high(integer)) div groupsize;
       bins[n]:=bins[n]+1;
       n:=random64(high(int64)) div groupsize64;
       bins64[n]:=bins64[n]+1;
     END;
     chsoneExt(bins,ebins,nbins,0,df,chsq,prob);
     chsoneExt(bins64,ebins64 ,nbins,0,df,chsq64,prob64);
     chsqsum:=chsqsum+chsq; probsum:=probsum+prob;
     chsqsum64:=chsqsum64+chsq64; probsum64:=probsum64+prob64;
     if j<=10 then
     begin
       add('');
       add(format('Trial # %d (%.0n randon samples}',[j,0.0+npts]));
       add('  32 bit    32 bit      64 bit    64 bit');
       add(' Expected  Observed    Expected  Observed');
       FOR i := 0 to nbins-1 DO
       BEGIN
          add(format('%7.0f  %7.0f      %7.0f %7.0f', [ebins[i],bins[i], ebins64[i], bins64[i]]));
       END;

       add('');
       add(format('DF=%.0f',[df]));
       add(format('Probability = Chance that drawing %d samples from the expected '
          + 'distribution would have a Chi-Sq value >= the observed Chi-Sq.',[npts]));
       add(format('32 bit Chi-squared: %10.4f,    Probability %10.4f',[chsq, prob]));
       add(format('64 bit Chi-Squared  %10.4f,    Probability %10.4f', [chsq64, prob64]));
     end;
   end;
   with memo1.Lines do
   begin
     add('');
     add(format('Mean values for %d trials with %.0n samples per trial',[nloops,0.0+npts]));
     add(format('32 bit Chi-squared: %10.4f,    Probability %10.4f',[chsqsum/nloops, probsum/nloops]));
     add(format('64 bit Chi-Squared  %10.4f,  Probability %10.4f', [chsqsum64/nloops, probsum64/nloops]));
   end;
   screen.cursor:=crdefault;
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
  ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;


end.
